import unittest
from nose.plugins.attrib import attr
import pycassa
from pycassa.cassandra.ttypes import NotFoundException
from agamemnon import DataStore

class GraphFunctionalTests(unittest.TestCase):
#This test requires a running cassandra instance at localhost:9160

    @attr(module='graph',type='functional')
    def test_graph_ops(self):
        system_manager = pycassa.system_manager.SystemManager()
        try:
            system_manager.drop_keyspace('my_test_cf')
        except NotFoundException:
            pass
        system_manager.create_keyspace('my_test_cf', replication_factor=1)
        #Create the Graph Datastore
        self.graph_db = DataStore('my_test_cf', pycassa.connect('my_test_cf', ['localhost:9160']),
                                      system_manager=pycassa.system_manager.SystemManager())
        #Create a spiderpig node
        self.graph_db.create_node('test_type', 'spiderpig', {'sound':'oink'})
        #get a spiderpig node (creating does return the node, but this is for testing and illustrative purposes
        spiderpig = self.graph_db.get_node('test_type', 'spiderpig')
        #make sure the attributes were set correctly
        self.failUnlessEqual(spiderpig['sound'], 'oink')
        #create a cow node
        self.graph_db.create_node('test_type', 'cow', {'sound':'moo'})
        #get the cow node
        cow = self.graph_db.get_node('test_type','cow')
        #test that the attributes were set correctly
        self.failUnlessEqual(cow['sound'], 'moo')
        #create a node of a different type
        self.graph_db.create_node('simpson', 'homer', {'sound':'Doh'})
        homer = self.graph_db.get_node('simpson', 'homer')
        self.failUnlessEqual(homer['sound'],'Doh')
        #create a relationship with a custom key and relationship attributes
        #this is a friend relationship
        spiderpig.friend(cow, key='spiderpig_cow_alliance', best=False)
        #get a reference node.  This is basically an index of all of the nodes of a given type
        #This returns the reference node for the test type
        reference_node = self.graph_db.get_reference_node('test_type')
        instances = ['spiderpig', 'cow']
        #Test to make sure that the correct instances are returned
        for rel in reference_node.instance:
            instances.remove(rel.target_node.key)
        #Get all of the outgoing friend relationships for spiderpig and make sure that cow is the only one
        for rel in spiderpig.friend.outgoing:
            self.failUnlessEqual(rel.key, 'spiderpig_cow_alliance')
            self.failUnlessEqual(rel.target_node.key,'cow')
        #Get all of the incoming friend relationships for cow and make sure that spiderpig is the only one
        for rel in cow.friend.incoming:
            self.failUnlessEqual(rel.key, 'spiderpig_cow_alliance')
            self.failUnlessEqual(rel.source_node.key, 'spiderpig')
            
        #Add Homer as a friend of spiderpig
        spiderpig.friend(homer, 'loves', AKA='Harry Plopper')
        for rel in homer.friend:
            self.failUnlessEqual(rel.source_node.key, 'spiderpig')

        self.failUnlessEqual(len(spiderpig.friend), 2)

        #Make sure that the "single" operation on the relationship type returns only one relationship
        self.failUnlessEqual(cow.friend.incoming.single.source_node.key, spiderpig.key)
        #delete the spiderpig node
        spiderpig.delete()
        #make sure that spiderpig is no logner in the cow's friend incoming list
        for rel in cow.friend.incoming:
            self.fail("shouldn't have rel: %s" % rel.key)
        for rel in homer.friend.incoming:
            self.fail("shouldn't have rel: %s" %rel.key)



# Copyright 2010 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'trhowe'


RELATIONSHIP_INDEX = '__relationship__index__'
OUTBOUND_RELATIONSHIP_CF = 'outbound__%s' % RELATIONSHIP_INDEX
INBOUND_RELATIONSHIP_CF = 'inbound__%s' % RELATIONSHIP_INDEX
RELATIONSHIP_KEY_PATTERN = '%s__%s'
ENDPOINT_NAME_TEMPLATE = '%s__%s'



from contextlib import nested
import unittest
from mock import Mock, patch
from nose.plugins.attrib import attr
import pycassa
from pycassa.cassandra.ttypes import NotFoundException
import agamemnon
import agamemnon.primitives as graph_prim
from agamemnon.graph_constants import ENDPOINT_NAME_TEMPLATE, RELATIONSHIP_KEY_PATTERN, RELATIONSHIP_INDEX

relationship_type = 'pig_cow_alliance'
relationship_types_cf = '__relationship__types'
type_cf = 'test_type'

class BatchMutatorMock(object):
    """
     because we are using the batch.Mutator context manager, we need to implement the magic methods
     __enter__ and __exit__ as they will be called to provide the contextualized object.
     contextualized_batch_mock is that object and is used for all of out assertions
    """

    def __init__(self):
        self._contextualized_batch_mock = Mock()
        self.__enter__ = Mock()
        self.__enter__.return_value = self._contextualized_batch_mock
        self.__exit__ = Mock()

    @property
    def contextualized_batch_mock(self):
        return self._contextualized_batch_mock

class DataStoreTests(unittest.TestCase):
    def setUp(self):
        self._pool = Mock()
        self._system_manager = Mock()
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            def side_effect(self, name):
                return name

            cf_mock.side_effect = side_effect
            self.store = agamemnon.DataStore('test_space', self._pool, self._system_manager)
        self.pig = graph_prim.Node(self.store, 'test_type', 'pig')
        self.cow = graph_prim.Node(self.store, 'test_type', 'cow')

    @attr(module='agamemnon', type='unit')
    def test_get_non_existent_cf(self):
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            cf_mock.side_effect = NotFoundException()
            self.failIf(self.store.cf_exists('test_type'))

    @attr(module='agamemnon', type='unit')
    def test_create_node(self):
        #create the node
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            def new_cf(*args):
                if new_cf.first:
                    new_cf.first = False
                    raise NotFoundException()
                else:
                    return cf_mock.return_value

            new_cf.first = True
            cf_mock.side_effect = new_cf

            def side_effect(name):
                if name is 'test_type':
                    return {'reference': 'reference'}
                else:
                    return Mock()

            cf_mock.return_value.get.side_effect = side_effect
            self.store.create_node('test_type', 'cow', {'sound': 'moo'})
            self._system_manager.create_column_family.assert_called_with('test_space', 'test_type', super=False,
                                                                         comparator_type=pycassa.system_manager.ASCII_TYPE)
            cf_mock.return_value.insert.assert_called_with('cow', {'sound': 'moo'})

    @attr(module='agamemnon', type='unit')
    def test_create_node_existing_cf(self):
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            def side_effect(name):
                if name is 'test_type':
                    return {'reference': 'reference'}
                else:
                    return Mock()

            cf_mock.return_value.get.side_effect = side_effect
            self.store.create_node('test_type', 'pig', {'sound': 'oink'})
            self.failIf(self._system_manager.create_column_family.called)

    @attr(module='agamemnon', type='unit')
    def test_create_duplicate_node(self):
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            def side_effect(name):
                if name is 'test_type':
                    return {'reference': 'reference'}
                else:
                    return Mock()

            cf_mock.return_value.get.side_effect = side_effect
            self.store.create_node('test_type', 'pig', {'sound': 'oink'})
            cf_mock.return_value.insert.assert_called_with('pig', {'sound': 'oink'})


    @attr(module='agamemnon', type='unit')
    def test_get_node(self):
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            cf_mock.return_value.get.return_value = {'sound': 'oink'}
            pig = self.store.get_node('test_type', 'pig')
            cf_mock.return_value.get.assert_called_with('pig')
            self.failUnlessEqual(pig.type, 'test_type')
            self.failUnlessEqual(pig.key, 'pig')
            self.failUnlessEqual(pig['sound'], 'oink')

    @attr(module='agamemnon', type='unit')
    def test_get_non_existent_node(self):
        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            #Test if node is not in datastore
            cf_mock.return_value.get.side_effect = NotFoundException()
            try:
                self.store.get_node('test_type', 'pig')
                self.fail()
            except agamemnon.NodeNotFoundException:
                pass

    @attr(module='agamemnon', type='unit')
    def test_delete_node_simple(self):
        node_mock = Mock()
        node_mock.remove = Mock()
        out_bound_rel_mock = Mock()
        out_bound_rel_mock.get.return_value = {}
        in_bound_rel_mock = Mock()
        in_bound_rel_mock.get.return_value = {}

        def side_effect(pool, type):
            if type == 'outbound__%s' % agamemnon.RELATIONSHIP_INDEX:
                return out_bound_rel_mock
            elif type == 'inbound__%s' % agamemnon.RELATIONSHIP_INDEX:
                return in_bound_rel_mock
            elif type == type_cf:
                return node_mock
            else:
                raise Exception('unknown cf: %s' % type)

        with nested(patch('pycassa.columnfamily.ColumnFamily'),
                    patch('pycassa.batch.Mutator')) as (cf_mock, batch_mock):
            batch_mock.return_value = BatchMutatorMock()
            contextualized_batch_mock = batch_mock.return_value.contextualized_batch_mock
            cf_mock.side_effect = side_effect
            pig = graph_prim.Node(self.store, 'test_type', 'pig')
            pig.delete()
            contextualized_batch_mock.remove.assert_called_with(node_mock, 'pig')


    @attr(module='agamemnon', type='unit')
    def test_create_relationship(self):
        # These are the inserts we expect to get.  The first one is adding friend to the list of relation types
        # the second and third are the to and from relationship inserts.  This is a pretty complicated action.  It
        # executes 7 inserts to create a relationship.  This could be cut down by eliminating the type specific
        # column families, but it is not clear that would make a significant difference in performance
        expected_calls = [
                (('outbound__%s' % RELATIONSHIP_INDEX, ENDPOINT_NAME_TEMPLATE % ('test_type', 'pig'),
                  {RELATIONSHIP_KEY_PATTERN % ('friend', 'pig_cow_alliance'): {'rel_type': 'friend',
                                                                               'rel_key': 'pig_cow_alliance',
                                                                               'source__type': 'test_type',
                                                                               'source__key': 'pig',
                                                                               'source__sound': 'oink',
                                                                               'target__type': 'test_type',
                                                                               'target__key': 'cow',
                                                                               'target__sound': 'moo'}}), {}),

                (('inbound__%s' % RELATIONSHIP_INDEX, ENDPOINT_NAME_TEMPLATE % ('test_type', 'cow'),
                  {'friend__pig_cow_alliance': {'rel_type': 'friend', 'rel_key': 'pig_cow_alliance',
                                                'source__type': 'test_type', 'source__key': 'pig',
                                                'source__sound': 'oink', 'target__type': 'test_type',
                                                'target__key': 'cow', 'target__sound': 'moo'}}), {})]

        pig = graph_prim.Node(self.store, 'test_type', 'pig', {'sound': 'oink'})
        cow = graph_prim.Node(self.store, 'test_type', 'cow', {'sound': 'moo'})
        with nested(patch('pycassa.columnfamily.ColumnFamily'),
                    patch('pycassa.batch.Mutator')) as (cf_mock, batch_mock):
            def side_effect(self, name):
                return name

            cf_mock.side_effect = side_effect

            batch_mock.return_value = BatchMutatorMock()
            contextualized_batch_mock = batch_mock.return_value.contextualized_batch_mock

            ####
            # This is the only method tested in this case.  Crazy.
            ####
            pig.friend(cow, key='pig_cow_alliance')

            # And, let's test our assumptions
            self.failUnlessEqual(contextualized_batch_mock.insert.call_args_list, expected_calls)

    @attr(module='agamemnon', type='unit')
    def test_delete_relationship(self):
        inbound_rel_index_cf = 'inbound__%s' % agamemnon.RELATIONSHIP_INDEX
        outbound_rel_index_cf = 'outbound__%s' % agamemnon.RELATIONSHIP_INDEX
        mocks = {inbound_rel_index_cf: Mock(), outbound_rel_index_cf: Mock()}
        from_key = ENDPOINT_NAME_TEMPLATE % (type_cf, 'pig')
        to_key = ENDPOINT_NAME_TEMPLATE % (type_cf, 'cow')
        relationship_key = 'friend__pig_cow_alliance'

        expected_calls = [((mocks[inbound_rel_index_cf], to_key), {'super_column': relationship_key}),
                          ((mocks[outbound_rel_index_cf], from_key), {'super_column': relationship_key})]

        with nested(patch('pycassa.columnfamily.ColumnFamily'),
                    patch('pycassa.batch.Mutator')) as (cf_mock, batch_mock):
            def side_effect(self, name):
                return mocks[name]

            cf_mock.side_effect = side_effect

            batch_mock.return_value = BatchMutatorMock()
            contextualized_batch_mock = batch_mock.return_value.contextualized_batch_mock
            rel = graph_prim.Relationship('pig_cow_alliance', graph_prim.Node(self.store, 'test_type', 'pig'),
                                          graph_prim.Node(self.store, 'test_type', 'cow'), self.store, 'friend')
            ####
            # Our method under test
            ####
            rel.delete()

            # Test our assumptions
            self.failUnlessEqual(contextualized_batch_mock.remove.call_args_list, expected_calls)

    @attr(module='agamemnon', type='unit')
    def test_get_relationships_of_given_type_outbound_only(self):
        inbound_rel_index_cf = 'inbound__%s' % agamemnon.RELATIONSHIP_INDEX
        outbound_rel_index_cf = 'outbound__%s' % agamemnon.RELATIONSHIP_INDEX
        out_bound_mock = Mock()
        out_bound_mock.get.return_value = {
            'friend__pig_cow_alliance': {'rel_type': 'friend', 'source__type': 'test_type', 'source_key': 'pig',
                                         'target__key': 'cow', 'target__type': 'test_type',
                                         'rel_id': 'pig_cow_alliance'}}
        in_bound_mock = Mock()
        in_bound_mock.get.side_effect = NotFoundException()
        pig = graph_prim.Node(self.store, 'test_type', 'pig')
        cow = graph_prim.Node(self.store, 'test_type', 'cow')

        def side_effect(pool, type):
            if type == outbound_rel_index_cf:
                return out_bound_mock
            elif type == inbound_rel_index_cf:
                return in_bound_mock
            else:
                raise Exception()

        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            cf_mock.side_effect = side_effect

            ####
            # Method under test
            ####
            for rel in pig.friend:
                expected = pig.friend(cow, 'pig_cow_alliance')
                self.failUnlessEqual(expected, rel)

    @attr(module='agamemnon', type='unit')
    def test_get_relationships_of_given_type_inbound_only(self):
        inbound_rel_index_cf = 'inbound__%s' % agamemnon.RELATIONSHIP_INDEX
        outbound_rel_index_cf = 'outbound__%s' % agamemnon.RELATIONSHIP_INDEX
        out_bound_mock = Mock()
        out_bound_mock.get.side_effect = NotFoundException()
        in_bound_mock = Mock()
        in_bound_mock.get.return_value = {
            'friend__chicken_pig_alliance': {'rel_type': 'friend', 'source__type': 'test_type', 'source_key': 'chicken',
                                             'target__key': 'pig', 'target__type': 'test_type',
                                             'rel_id': 'chicken_pig_alliance'}}
        chicken = graph_prim.Node(self.store, 'test_type', 'chicken')
        pig = graph_prim.Node(self.store, 'test_type', 'pig')

        def side_effect(pool, type):
            if type == outbound_rel_index_cf:
                return out_bound_mock
            elif type == inbound_rel_index_cf:
                return in_bound_mock
            else:
                raise Exception()

        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
            cf_mock.side_effect = side_effect

            ####
            # Method under test
            ####
            for rel in pig.friend:
                expected = chicken.friend(pig, 'chicken_pig_alliance')
                self.failUnlessEqual(expected, rel)

    @attr(module='agamemnon', type='unit')
    def test_delete_node_with_relationships(self):
        mocks = {'test_type': Mock(), 'outbound__%s' % agamemnon.RELATIONSHIP_INDEX: Mock(),
                 'inbound__%s' % agamemnon.RELATIONSHIP_INDEX: Mock()}
        inbound = 'inbound__%s' % agamemnon.RELATIONSHIP_INDEX
        outbound = 'outbound__%s' % agamemnon.RELATIONSHIP_INDEX
        mocks['test_type'].remove = Mock()
        mocks[inbound].get.side_effect = NotFoundException()

        mocks[outbound].get.return_value = {
            'friend__pig_cow_alliance': {'rel_type': 'friend', 'rel_key': 'pig_cow_alliance',
                                         'source__type': 'test_type', 'source__key': 'pig',
                                         'source__sound': 'oink', 'target__type': 'test_type',
                                         'target__key': 'cow', 'target__sound': 'moo'}}
        from_key = ENDPOINT_NAME_TEMPLATE % ('test_type', 'pig')
        to_key = ENDPOINT_NAME_TEMPLATE % ('test_type', 'cow')
        relationship_key = 'friend__pig_cow_alliance'

        expected_calls = [((mocks[inbound], to_key), {'super_column': relationship_key}),
                          ((mocks[outbound], from_key), {}),
                          ((mocks[inbound], from_key), {}),
                          ((mocks['test_type'], 'pig'), {})]


        def side_effect(pool, type):
            return mocks[type]

        with nested(patch('pycassa.columnfamily.ColumnFamily'),
                    patch('pycassa.batch.Mutator')) as(cf_mock, batch_mock):
            batch_mock.return_value = BatchMutatorMock()
            contextual_batch_mock = batch_mock.return_value.contextualized_batch_mock
            cf_mock.side_effect = side_effect
            self.pig.delete()
            call_args_list = contextual_batch_mock.remove.call_args_list
            self.failUnlessEqual(call_args_list, expected_calls)






# Copyright 2010 University of Chicago
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
import logging
import pycassa.batch as batch
from pycassa.cassandra.ttypes import NotFoundException
import pycassa.system_manager as system_manager
import pycassa.columnfamily as cf
import agamemnon.primitives as prim
from graph_constants import *

log = logging.getLogger(__name__)

class NoTransactionError(Exception):
    pass


class RelationshipIndexEntry(object):
    def __init__(self, cf, key, rel_key):
        self._entry = {rel_key: {'__cf': cf, '__key': key, '__rel': rel_key}}

    @property
    def entry(self):
        return self._entry


class DataStore(object):
    def __init__(self, keyspace, pool, system_manager):
        self._cf_cache = {}
        self._index_cache = {}
        self._system_manager = system_manager
        self._pool = pool
        self._keyspace = keyspace
        self.outbound_rel_index_cf = 'outbound__%s' % RELATIONSHIP_INDEX
        self.inbound_rel_index_cf = 'inbound__%s' % RELATIONSHIP_INDEX
        if not self.cf_exists(self.outbound_rel_index_cf):
            self.create_cf(self.outbound_rel_index_cf, super=True)
        if not self.cf_exists(self.inbound_rel_index_cf):
            self.create_cf(self.inbound_rel_index_cf, super=True)

    def create_cf(self, type, column_type=system_manager.ASCII_TYPE, super=False, index_columns=list()):
        self._system_manager.create_column_family(self._keyspace, type, super=super, comparator_type=column_type)
        for column in index_columns:
            self._system_manager.create_index(self._keyspace, type, column, column_type,
                                              index_name='%s_%s_index' % (type, column))

    def cf_exists(self, type):
        try:
            cf.ColumnFamily(self._pool, type)
        except NotFoundException:
            return False
        return True

    def get_cf(self, type):
        if type in self._cf_cache.keys():
            return self._cf_cache[type]
        else:
            try:
                column_family = cf.ColumnFamily(self._pool, type)
                self._cf_cache[type] = column_family
            except NotFoundException:
                return None
            return column_family

    def get(self, type, row_key, super_column_key=None):
        column_family = self.get_cf(type)
        if column_family is None:
            self.create_cf(type)
            column_family = self.get_cf(type)
        if super_column_key is None:
            values = column_family.get(row_key)
        else:
            values = column_family.get(row_key, super_column=super_column_key)
        return values

    def delete(self, type, key, super_key=None):
        if super_key is None:
            self.get_cf(type).remove(key)
        else:
            self.get_cf(type).remove(key, super_key)

    def insert(self, type, key, args, super_key=None):
        if not self.cf_exists(type):
            self.create_cf(type)
        column_family = self.get_cf(type)
        if super_key is None:
            column_family.insert(key, args)
        else:
            column_family.insert(key, {super_key: args})

    def get_outgoing_relationships(self, source_node, rel_type, count=100):
        source_key = RELATIONSHIP_KEY_PATTERN % (source_node.type, source_node.key)
        cf = self.get_cf(OUTBOUND_RELATIONSHIP_CF)
        #Ok, this is weird.  So in order to get a column slice, you need to provide a start that is <= your first column
        #id, and a finish which is >= your last column.  Since our columns are sorted by ascii, this means we need to go
        #from rel_type_ to rel_type` because "`" is the char 1 greater than "_", so this will get anything which starts
        #rel_type_.  Now I realize that this could problems when there is a "_" in the relationship name, so we will
        #probably need a different delimiter.
        #TODO: fix delimiter
        try:
            super_columns = cf.get(source_key, column_start='%s__' % rel_type, column_finish='%s_`' % rel_type, column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_outgoing_relationship(rel_type, source_node, super_column) for super_column in
                super_columns.items()]


    def get_incoming_relationships(self, target_node, rel_type, count=100):
        target_key = RELATIONSHIP_KEY_PATTERN % (target_node.type, target_node.key)
        cf = self.get_cf(INBOUND_RELATIONSHIP_CF)
        #Ok, this is weird.  So in order to get a column slice, you need to provide a start that is <= your first column
        #id, and a finish which is >= your last column.  Since our columns are sorted by ascii, this means we need to go
        #from rel_type_ to rel_type` because "`" is the char 1 greater than "_", so this will get anything which starts
        #rel_type_.  Now I realize that this could problems when there is a "_" in the relationship name, so we will
        #probably need a different delimiter.
        #TODO: fix delimiter
        try:
            super_columns = cf.get(target_key, column_start='%s__' % rel_type, column_finish='%s_`' % rel_type, column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_incoming_relationship(rel_type, target_node, super_column) for super_column in
                super_columns.items()]

    def get_outgoing_relationship(self, rel_type, source_node, super_column):
        """
        Process the contents of a SuperColumn to extract the relationship and to_node properties and return
        a constructed relationship
        """
        rel_key = super_column[0]
        target_node_key = None
        target_node_type = None
        target_attributes = {}
        rel_attributes = {}
        for column in super_column[1].keys():
            value = super_column[1][column]
            if column == 'target__type':
                target_node_type = value
            elif column == 'target__key':
                target_node_key = value
            elif column.startswith('target__'):
                target_attributes[column[8:]] = value
            else:
                rel_attributes[column] = value
        return prim.Relationship(rel_key, source_node,
                                 prim.Node(self, target_node_type, target_node_key, target_attributes), self
                                 , rel_type,
                                 rel_attributes)

    def get_incoming_relationship(self, rel_type, target_node, super_column):
        """
        Process the contents of a SuperColumn to extract an incoming relationship and the associated from_node and
        return a constructed relationship
        """
        rel_key = super_column[0]
        source_node_key = None
        source_node_type = None
        source_attributes = {}
        rel_attributes = {}
        for column in super_column[1].keys():
            value = super_column[1][column]
            if column == 'source__type':
                source_node_type = value
            elif column == 'source__key':
                source_node_key = value
            elif column.startswith('source__'):
                source_attributes[column[8:]] = value
            else:
                rel_attributes[column] = value
        return prim.Relationship(rel_key, prim.Node(self, source_node_type, source_node_key, source_attributes),
                                 target_node,
                                 self, rel_type, rel_attributes)


    def delete_relationship(self, rel_type, rel_id, from_type, from_key, to_type, to_key):
        from_key = ENDPOINT_NAME_TEMPLATE % (from_type, from_key)
        to_key = ENDPOINT_NAME_TEMPLATE % (to_type, to_key)

        with batch.Mutator(self._pool) as b:
            b.remove(self.get_cf(INBOUND_RELATIONSHIP_CF), to_key,
                     super_column=RELATIONSHIP_KEY_PATTERN % (rel_type, rel_id))
            b.remove(self.get_cf(OUTBOUND_RELATIONSHIP_CF), from_key,
                     super_column=RELATIONSHIP_KEY_PATTERN % (rel_type, rel_id))

    def create_relationship(self, rel_type, source_node, target_node, key, args):
        if key is None:
            rel_key = RELATIONSHIP_KEY_PATTERN % (rel_type, datetime.now())
        else:
            rel_key = RELATIONSHIP_KEY_PATTERN % (rel_type, key)
            #node relationship types
        with batch.Mutator(self._pool) as b:
            #relationship attributes
            columns = {}
            columns.update(args)

            #outbound_cf
            columns = {'rel_type': rel_type, 'rel_key': key}
            #add relationship attributes
            columns.update(columns)
            #add target attributes
            columns['target__type'] = target_node.type.encode('ascii')
            columns['target__key'] = target_node.key.encode('ascii')
            target_attributes = target_node.attributes
            for attribute_key in target_attributes.keys():
                columns['target__%s' % attribute_key] = target_attributes[attribute_key]
            columns['source__type'] = source_node.type.encode('ascii')
            columns['source__key'] = source_node.key.encode('ascii')
            source_attributes = source_node.attributes
            for attribute_key in source_attributes.keys():
                columns['source__%s' % attribute_key] = source_attributes[attribute_key]

            source_key = ENDPOINT_NAME_TEMPLATE % (source_node.type, source_node.key)
            target_key = ENDPOINT_NAME_TEMPLATE % (target_node.type, target_node.key)
            inbound_rel_index_cf = self.get_cf(INBOUND_RELATIONSHIP_CF)
            outbound_rel_index_cf = self.get_cf(OUTBOUND_RELATIONSHIP_CF)

            b.insert(outbound_rel_index_cf, source_key, {rel_key: columns})
            b.insert(inbound_rel_index_cf, target_key, {rel_key: columns})

        #created relationship object
        return prim.Relationship(rel_key, source_node, target_node, self, rel_type, args)

    def create_node(self, type, key, args=None, reference=False):
        node = prim.Node(self, type, key, args)
        self.insert(type, key, args)
        if not reference:
            #this adds the created node to the reference node for this type of object
            #that reference node functions as an index to easily access all nodes of a specific type
            reference_node = self.get_reference_node(type)
            reference_node.instance(node, key=key)
        return node

    def delete_node(self, node):
        node_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
        try:
            outbound_results = self.get(OUTBOUND_RELATIONSHIP_CF, node_key)
        except NotFoundException:
            outbound_results = {}
        try:
            inbound_results = self.get(INBOUND_RELATIONSHIP_CF, node_key)
        except NotFoundException:
            inbound_results = {}

        with batch.Mutator(self._pool) as b:
            for key in outbound_results.keys():
                rel = outbound_results[key]
                b.remove(self.get_cf(INBOUND_RELATIONSHIP_CF),
                         RELATIONSHIP_KEY_PATTERN % (rel['target__type'], rel['target__key']),
                         super_column='%s__%s' % (rel['rel_type'], rel['rel_key']))
            for key in inbound_results.keys():
                rel = inbound_results[key]
                b.remove(self.get_cf(OUTBOUND_RELATIONSHIP_CF),
                         RELATIONSHIP_KEY_PATTERN % (rel['source__type'], rel['source__key']),
                         super_column='%s__%s' % (rel['rel_type'], rel['rel_key']))
            b.remove(self.get_cf(OUTBOUND_RELATIONSHIP_CF), node_key)
            b.remove(self.get_cf(INBOUND_RELATIONSHIP_CF), node_key)
            b.remove(self.get_cf(node.type), node.key)

    def save_node(self, node):
        """
        This needs to update the entry in the type table as well as all of the relationships
        """
        source_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
        target_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)

        try:
            outbound_results = self.get(OUTBOUND_RELATIONSHIP_CF, source_key)
        except NotFoundException:
            outbound_results = {}
        try:
            inbound_results = self.get(INBOUND_RELATIONSHIP_CF, target_key)
        except NotFoundException:
            inbound_results = {}
        outbound_columns = {'source__type': node.type.encode('utf-8'), 'source__key': node.key.encode('utf-8')}
        node_attributes = node.attributes
        for attribute_key in node.attributes.keys():
            outbound_columns['source__%s' % attribute_key] = node_attributes[attribute_key]
        for key in outbound_results.keys():
            target = outbound_results[key]
            target_key = ENDPOINT_NAME_TEMPLATE %(target['target__type'], target['target__key'])
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, outbound_columns, key)
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, outbound_columns, key)
        inbound_columns = {'target__type': node.type.encode('utf-8'), 'target__key': node.key.encode('utf-8')}
        for attribute_key in node.attributes.keys():
            inbound_columns['target__%s' % attribute_key] = node_attributes[attribute_key]
        for key in inbound_results.keys():
            source = inbound_results[key]
            source_key = ENDPOINT_NAME_TEMPLATE %(source['source__type'], source['source__key'])
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, inbound_columns, key)
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, inbound_columns, key)
        self.insert(node.type, node.key, node.attributes)

    def get_node(self, type, key):
        try:
            values = self.get(type, key)
        except NotFoundException:
            raise NodeNotFoundException()
        return prim.Node(self, type, key, values)

    def get_reference_node(self, name):
        """
        Nodes returned here are very easily referenced by name and then function as an index for all attached nodes
        The most typical use case is to index all of the nodes of a certain type, but the functionality is not limited
        to this.  
        """
        try:
            node = self.get_node('reference', name)
        except NodeNotFoundException:
            node = self.create_node('reference', name, {'reference': 'reference'}, reference=True)
        return node


class NodeNotFoundException(Exception):
    pass


def _get_args(args, **kwargs):
    tmp_args = args
    if tmp_args is not None:
        if kwargs is not None:
            tmp_args.update(kwargs)
    else:
        if kwargs is not None:
            tmp_args = kwargs
    return tmp_args


def DFS(node, relationship_type, return_predicate=None):
    visited = set([node.key])
    S = [relationship for relationship in getattr(node, relationship_type)]
    while S:
        p = S.pop()
        relationship = p
        child = relationship.target_node
        if child.key not in visited:
            if return_predicate is not None and return_predicate(relationship, child):
                visited.add(child.key)
                yield child
            elif return_predicate is None:
                visited.add(child.key)
                yield child
            if hasattr(child, relationship_type):
                visited.add(child.key)
                S.extend([relationship for relationship in getattr(child, relationship_type)])



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open('README.rst').read()
CHANGES = open('CHANGES.txt').read()

requires = ['pycassa', 'thrift05', 'setuptools']
tests_requires = requires + ['nose', 'mock']

setup(name='agamemnon',
      version='0.1.3.1',
      description='A graph database built on top of cassandra',
      long_description=README + "\n\n" + CHANGES,
      classifiers=[
        "Programming Language :: Python",
        ],
      author='Tom Howe',
      author_email='trhowe@ci.uchicago.edu',
      url='https://github.com/turtlebender/agamemnon',
      license='LICENSE.txt',
      keywords='cassandra',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=True,
      setup_requires=['nose>=0.11'],
      install_requires=requires,
      tests_require=tests_requires,
      test_suite="nose.collector",
      )

